
# Aspidites
# Copyright (C) 2021 Ross J. Duff

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from ._vendor.contracts import check, contract, new_contract


@new_contract
@contract
def heritable(bases: 'tuple') -> 'bool':
    """Convenience clause for a tuple of heritable bases"""
    return not bool(len(tuple(b for b in bases if isinstance(b, _Final))))


# noinspection PyPep8Naming
class _Final(type):
    """Non public metaclass implementation for final classes"""
    def __new__(mcs, name, bases, classdict):
        check('heritable', bases)
        return type.__new__(mcs, name, bases, dict(classdict))


def final(_=_Final):  # This is 100% a hack: but it works.
    """Decorator to create final classes like:
    .. code:: python

        @final()
        class Foo(object):
            ...
    """
    if _ != _Final:
        meta = _Final
    else:
        meta = _

    def wrapper(mcs):
        __name = str(mcs.__name__)
        __bases = tuple(mcs.__bases__)
        __dict = dict(mcs.__dict__)

        for each_slot in __dict.get("__slots__", tuple()):
            __dict.pop(each_slot, None)

        __dict["__metaclass__"] = meta
        __dict["__wrapped__"] = mcs

        return meta(__name, __bases, __dict)
    return wrapper



from ..interface import ExternalScopedVariableNotFound
from ..syntax import S, W
from ..utils import ignore_typeerror
from pyparsing import Word, alphanums
import inspect
from Aspidites._vendor.contracts.library.types_misc import CheckType


def _lookup_from_calling_scope(token):
    """
    Extract the value of the token from the scope where the spec is defined
    """

    # We walk the callstack from the outside in, searching for the
    # frame where the spec is defined
    #
    # XXX Check if there are other places where a spec might be defined
    from .. import decorate, parse, check, fail

    def _code(f):
        try:  # pragma: no cover
            return f.func_code
        except AttributeError:  # py3
            return f.__code__

    frames = inspect.getouterframes(inspect.currentframe())
    frames = [f[0] for f in frames[::-1]]
    fcodes = [f.f_code for f in frames]

    def find_invokation(func):
        # return the first frame where func is called, or raise ValueError
        # find first frame inside function, step out 1
        return lambda: frames[fcodes.index(_code(func)) - 1]

    def find_decorate():
        # return the first frame where decorate is called, or raise ValueError

        # Brittle: We must to determine whether user calls decorate
        #          directly (in which case relevant scope is 1 frame out)
        #          or indirectly via @contract() (scope is 2 frames out)
        #          The implementation relies on the name `tmp_wrap` of the
        #          hidden function inside @contract.
        idx = fcodes.index(_code(decorate))

        if frames[idx - 1].f_code.co_name == 'tmp_wrap':
            # decorate() called via @contract, Step out 2 frames
            return frames[idx - 2]

        # decorate() called oustide of @contract, step out 1 frame
        return frames[idx - 1]

    # search order important
    searchers = [find_decorate,
                 find_invokation(check),
                 find_invokation(fail),
                 find_invokation(parse)]
    for s in searchers:
        try:
            f = s()
        except (ValueError, IndexError):
            continue
        if not f:
            continue
        try:
            return eval(token, f.f_locals, f.f_globals)
        except NameError:
            raise ExternalScopedVariableNotFound(token)

    raise RuntimeError("Cound not find a scope to lookup %s" % token)


# class ScopedVariableRef(RValue):
#
#     """
#     A variable whose value is extracted by name from the scope where the spec is defined.
#     """
#
#     def __init__(self, value, where=None):
#         self.where = where
#         self.value = value
#
#     def eval(self, context):
#         return self.value
#
#     def __repr__(self):
#         return "ScopedVariableRef(%r)" % self.value
#
#     def __str__(self):
#         return str(self.value)

@ignore_typeerror
def scoped_parse_action(s, loc, tokens):
    assert len(tokens) == 1
    where = W(s, loc)
    val = _lookup_from_calling_scope(tokens[0])

    from Aspidites._vendor.contracts.library.simple_values import SimpleRValue

    from Aspidites._vendor.contracts.inspection import can_be_used_as_a_type

    if can_be_used_as_a_type(val):
        return CheckType(val)
    else:
        return SimpleRValue(value=val, where=where, representation=s)

scoped_variables = (S('$') + Word(alphanums + '_'))
scoped_variables_ref = scoped_variables.setParseAction(scoped_parse_action)








from .._compat import IS_PY2 as IS_PY2
from .compile import compile_restricted_eval as compile_restricted_eval
from typing import Any

nltosp: Any
default_guarded_getattr = getattr

def default_guarded_getitem(ob, index): ...
def default_guarded_getiter(ob): ...

class RestrictionCapableEval:
    globals: Any
    rcode: Any
    ucode: Any
    used: Any
    __name__: Any
    expr: Any
    def __init__(self, expr) -> None: ...
    def prepRestrictedCode(self) -> None: ...
    def prepUnrestrictedCode(self) -> None: ...
    def eval(self, mapping): ...
    def __call__(self, **kw): ...



from Aspidites._vendor.RestrictedPython.transformer import RestrictingNodeTransformer as RestrictingNodeTransformer
from Aspidites._vendor._compat import IS_CPYTHON as IS_CPYTHON, IS_PY2 as IS_PY2
from typing import Any, NamedTuple

class CompileResult(NamedTuple):
    code: Any
    errors: Any
    warnings: Any
    used_names: Any
syntax_error_template: str
NOT_CPYTHON_WARNING: str

def compile_restricted_exec(source, filename: str = ..., flags: int = ..., dont_inherit: bool = ..., policy=...): ...
def compile_restricted_eval(source, filename: str = ..., flags: int = ..., dont_inherit: bool = ..., policy=...): ...
def compile_restricted_single(source, filename: str = ..., flags: int = ..., dont_inherit: bool = ..., policy=...): ...
def compile_restricted_function(p, body, name, filename: str = ..., globalize: Any | None = ..., flags: int = ..., dont_inherit: bool = ..., policy=...): ...
def compile_restricted(source, filename: str = ..., mode: str = ..., flags: int = ..., dont_inherit: bool = ..., policy=...): ...



from . import good, fail

class BaseClass():
    pass

class SubClass(BaseClass):
    pass

good('isinstance(BaseClass)', BaseClass())
good('isinstance(BaseClass)', SubClass())
good('isinstance(SubClass)', SubClass())
fail('isinstance(SubClass)', BaseClass())

class BaseClass2(object):
    pass

class SubClass2(BaseClass2):
    pass

good('isinstance(BaseClass2)', BaseClass2())
good('isinstance(BaseClass2)', SubClass2())
good('isinstance(SubClass2)', SubClass2())
fail('isinstance(SubClass2)', BaseClass2())

# Let's try with 3, old-style

class BaseClass3():
    pass

class MidClass3(BaseClass3):
    pass

class SubClass3(MidClass3):
    pass

good('isinstance(BaseClass3)', BaseClass3())
good('isinstance(BaseClass3)', SubClass3())
good('isinstance(SubClass3)', SubClass3())
fail('isinstance(SubClass3)', BaseClass3())

class BaseClass4(object):
    pass

class MidClass4(BaseClass4):
    pass

class SubClass4(MidClass4):
    pass

good('isinstance(BaseClass4)', BaseClass4())
good('isinstance(BaseClass4)', SubClass4())
good('isinstance(SubClass4)', SubClass4())
fail('isinstance(SubClass4)', BaseClass4())




##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""Restricted Python Expressions."""

from .._compat import IS_PY2
from .compile import compile_restricted_eval

import ast


if IS_PY2:  # pragma: no cover
    from string import maketrans
else:  # pragma: PY3
    maketrans = str.maketrans


nltosp = maketrans('\r\n', '  ')

# No restrictions.
default_guarded_getattr = getattr


def default_guarded_getitem(ob, index):
    # No restrictions.
    return ob[index]


def default_guarded_getiter(ob):
    # No restrictions.
    return ob


class RestrictionCapableEval(object):
    """A base class for restricted code."""

    globals = {'__builtins__': None}
    # restricted
    rcode = None

    # unrestricted
    ucode = None

    # Names used by the expression
    used = None

    def __init__(self, expr):
        """Create a restricted expression

        where:

          expr -- a string containing the expression to be evaluated.
        """
        expr = expr.strip()
        self.__name__ = expr
        expr = expr.translate(nltosp)
        self.expr = expr
        # Catch syntax errors.
        self.prepUnrestrictedCode()

    def prepRestrictedCode(self):
        if self.rcode is None:
            result = compile_restricted_eval(self.expr, '<string>')
            if result.errors:
                raise SyntaxError(result.errors[0])
            self.used = tuple(result.used_names)
            self.rcode = result.code

    def prepUnrestrictedCode(self):
        if self.ucode is None:
            exp_node = compile(
                self.expr,
                '<string>',
                'eval',
                ast.PyCF_ONLY_AST)

            co = compile(exp_node, '<string>', 'eval')

            # Examine the ast to discover which names the expression needs.
            if self.used is None:
                used = set()
                for node in ast.walk(exp_node):
                    if isinstance(node, ast.Name):
                        if isinstance(node.ctx, ast.Load):
                            used.add(node.id)

                self.used = tuple(used)

            self.ucode = co

    def eval(self, mapping):
        # This default implementation is probably not very useful. :-(
        # This is meant to be overridden.
        self.prepRestrictedCode()

        global_scope = {
            '_getattr_': default_guarded_getattr,
            '_getitem_': default_guarded_getitem,
            '_getiter_': default_guarded_getiter,
        }

        global_scope.update(self.globals)

        for name in self.used:
            if (name not in global_scope) and (name in mapping):
                global_scope[name] = mapping[name]

        return eval(self.rcode, global_scope)

    def __call__(self, **kw):
        return self.eval(kw)



from Aspidites._vendor.RestrictedPython import compile_restricted_eval
from Aspidites._vendor._compat import IS_PY35_OR_GREATER
from Aspidites.tests._vendor.RestrictedPython.helper import restricted_eval
import pytest


# Arithmetic Operators

def test_Add():
    assert restricted_eval('1 + 1') == 2


def test_Sub():
    assert restricted_eval('5 - 3') == 2


def test_Mult():
    assert restricted_eval('2 * 2') == 4


def test_Div():
    assert restricted_eval('10 / 2') == 5


def test_Mod():
    assert restricted_eval('10 % 3') == 1


def test_Pow():
    assert restricted_eval('2 ** 8') == 256


def test_FloorDiv():
    assert restricted_eval('7 // 2') == 3


@pytest.mark.skipif(
    not IS_PY35_OR_GREATER,
    reason="MatMult was introducted on Python 3.5")
def test_MatMult():
    result = compile_restricted_eval('(8, 3, 5) @ (2, 7, 1)')
    assert result.errors == (
        'Line None: MatMult statements are not allowed.',
    )
    assert result.code is None



#!/usr/bin/python3
import sys
from pathlib import Path
from setuptools import setup, find_packages
from setuptools.dist import Distribution
from setuptools.command.install import install
# ~#~ # Build static libs # ~#~ #
from Cython.Build import cythonize
ext_modules = cythonize([str(Path('Aspidites/monads.py')), str(Path('Aspidites/math.py'))])
from Aspidites import __version__, __license__, __title__, __author__, compiler, parser
from Aspidites.__main__ import get_cy_kwargs
cy_kwargs = get_cy_kwargs()
cy_kwargs.update(embed="'main'")
code = open(Path('Aspidites/woma/library.wom'), 'r').read()
cy_kwargs.update(
    code=parser.parse_module(code),
    force=True,
    fname='Aspidites/woma/library.pyx',
    bytecode=True,
    c=True,
    verbose=0,
    build_requires=''
)
compiler.Compiler(**cy_kwargs)


def read(fname):
    return open(Path(fname).absolute()).read()


# Tested with wheel v0.29.0, 0.36.2
class BinaryDistribution(Distribution):
    """Distribution which always forces a binary package with platform name"""
    def has_ext_modules(foo):
        return True


class InstallWrapper(install):
    """Provides a install wrapper for native woma
    extensions. These don't really belong in the
    Python package."""

    def run(self):
        # Run this first so the install stops in case
        # these fail otherwise the Python package is
        # successfully installed
        print("running preinstall hooks")
        self.preinstall()
        install.run(self)  # pip install
        print("running postinstall hooks")
        self.postinstall()

    def preinstall(self):
        """preinstall hook"""
        # target = Path('build/lib/Aspidites/woma/library.wom')
        # output = Path('build/lib/Aspidites/woma/library.pyx')
        # c = "Aspidites %s -c -o %s --embed=True" % (target, output)
        # os.popen(c)
        # print(c)
        pass

    def postinstall(self):
        """postinstall hook"""
        pass


setup(
    name=__title__,
    version=__version__,
    author=__author__,
    author_email="rjdbcm@mail.umkc.edu",
    description="Aspidites is the reference implementation of the Woma Language",
    license=__license__,
    keywords="language",
    url="https://github.com/rjdbcm/Aspidites",
    install_requires=[
        'pyrsistent',
        'numpy',
        'cython>0.28,<3',
        'pyparsing',
        'mypy',
        'pytest',
        'pytest-xdist',
        'pytest-mock',
        'hypothesis',
        'future'
        ],
    packages=find_packages(),
    ext_modules=ext_modules,
    test_suite='Aspidites/tests',
    distclass=Distribution if sys.platform != 'darwin' else BinaryDistribution,
    entry_points={'console_scripts': ['aspidites = Aspidites.__main__:main']},
    package_data={'': ["*.wom", "*.pyx", "*.pyi", "*.so", "*.c"]},  # add any native *.wom files
    long_description=read('README.md'),
    cmdclass={'install': InstallWrapper},
    long_description_content_type='text/markdown',
    classifiers=[
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Programming Language :: Other",
        # "Programming Language :: Python :: 3.6", EOL in December 2021 and don't want to vendor dataclasses
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        # "Programming Language :: Python :: Implementation :: PyPy", looks like a no go
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Software Development :: Compilers",
        "Topic :: Utilities",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "License :: OSI Approved :: MIT License",
        "License :: OSI Approved :: BSD License",
        # "License :: OSI Approved :: Zope Public License (ZPL)", ??? Invalid
        "License :: OSI Approved :: Apache Software License",
    ],
)
